@@ -11,6 +11,7 @@ from typing import Any, Optional, Union, Generator, List
 from urllib.parse import urlparse
 import tldextract
 from ftfy import fix_text
+import sqlite3
 
 import pandas as pd
 from .exceptions import FileError, EncodingError, ValidationError
@@ -76,6 +77,68 @@ def standardized_string(string: Optional[str] = None) -> str:
     except:
         return old_string
 
+def _replace_in_string(text, old, new, count, ignore_case, position):
+    """Helper function to replace text safely, with optional position occurrence."""
+    if ignore_case:
+        pattern = re.compile(re.escape(old), re.IGNORECASE)
+    else:
+        pattern = re.compile(re.escape(old))
+
+    # Replace only position occurrence
+    if position is not None:
+        matches = list(pattern.finditer(text))
+        if len(matches) < position:
+            return text  # position occurrence doesn't exist
+        start, end = matches[position - 1].span()
+        return text[:start] + new + text[end:]
+
+    # Replace first 'count' occurrences (or all if count=None)
+    if count is None:
+        return pattern.sub(new, text)
+    else:
+        return pattern.sub(new, text, count)
+
+def replace_content(content, old="", new="", count=None, ignore_case=False, position=None):
+    """
+    Safely replace content in a string or file.
+
+    Parameters:
+        content     : str or path to a file
+        old         : substring to replace (default: "")
+        new         : replacement string (default: "")
+        count       : max number of occurrences to replace (like str.replace)
+        ignore_case : bool - whether to ignore case
+        position         : int - replace only the position occurrence (1-based). If None, replace all/count
+
+    Returns:
+        str : replaced text (or original content if replacement fails)
+    """
+    try:
+        if old == "":
+            print("Warning: 'old' is empty — no replacement performed.")
+            return content
+
+        # Read file if content is a path to a file
+        if isinstance(content, str) and content.endswith(('.txt', '.md', '.log')):
+            with open(content, 'r', encoding='utf-8') as f:
+                text = f.read()
+            text = _replace_in_string(text, old, new, count, ignore_case, position)
+            with open(content, 'w', encoding='utf-8') as f:
+                f.write(text)
+            return text
+
+        # String content
+        elif isinstance(content, str):
+            return _replace_in_string(content, old, new, count, ignore_case, position)
+
+        # Convert other types to string
+        else:
+            text = str(content)
+            return _replace_in_string(text, old, new, count, ignore_case, position)
+
+    except Exception as e:
+        print(f"Error replacing content: {type(e).__name__} - {e}")
+        return content
 
 # -------------------------------
 # Input Validation
@@ -117,6 +180,33 @@ def detect_encoding(data: Union[str, bytes]) -> str:
     except (UnicodeDecodeError, LookupError):
         return 'utf-8'
 
+# -------------------------------
+# Xpath detection
+# -------------------------------
+
+def _detect_selector_method(selector: str) -> str:
+    """
+    Detect whether the selector is XPath or CSS with more robust rules.
+    """
+    selector = selector.strip()
+
+    # Strong XPath patterns
+    xpath_patterns = [
+        r"(^//|^\.//)",         # Starts with '//' or './/'
+        r"@",                   # Attribute access
+        r"\bcontains\(",        # XPath function
+        r"\bstarts-with\(",     # XPath function
+        r"text\(\)",            # XPath text node
+        r"::",                  # Axis specifier
+    ]
+
+    if any(re.search(pattern, selector) for pattern in xpath_patterns):
+        return "xpath"
+
+    # Default fallback → CSS
+    return "css"
+        
+
 def get_base_domain(url: str) -> Optional[str]:
     """
     Extracts the registered base domain from a given URL using tldextract.
@@ -319,6 +409,103 @@ def save_to_xls(data_list: Optional[List[list]] = None, headers: Optional[List[s
         print(f"Data not saved due to missing arguments: {', '.join(missing_args)}")
 
 
+def save_to_db(data_list: Optional[List[list]] = None, headers: Optional[List[str]] = None, db_file_path: str = "data.sqlite", table_name: str = "data", auto_data_type: bool = False) -> None:
+    """
+    Saves data to an SQLite database. If the table exists, it appends the data; otherwise, it creates a new table.
+    Handles type detection, smart appending, automatically adds missing columns with default empty string values.
+
+    Args:
+    - data_list (Optional[List[list]]): The data to be saved in the database.
+    - headers (Optional[List[str]]): The column headers for the table.
+    - db_file_path (str, optional): The path to the SQLite database file. Defaults to "data.sqlite".
+    - table_name (str, optional): The name of the table to save data. Defaults to "data".
+    - auto_data_type (bool, optional): If True, detect column types automatically. Defaults to False.
+
+    Returns:
+    - None
+    """
+    if data_list is None or headers is None:
+        missing_args = []
+        if data_list is None:
+            missing_args.append('data_list')
+        if headers is None:
+            missing_args.append('headers')
+        print(f"Data not saved due to missing arguments: {', '.join(missing_args)}")
+        return
+
+    try:
+        # Ensure the directory for the database exists
+        dir_name = get_dir_by_path(file_path=db_file_path)
+        create_directory(directory_name=dir_name)
+
+        conn = sqlite3.connect(db_file_path)
+        cursor = conn.cursor()
+
+        def detect_type(value):
+            if isinstance(value, int):
+                return "INTEGER"
+            elif isinstance(value, float):
+                return "REAL"
+            else:
+                return "TEXT"
+
+        # Check if table exists
+        cursor.execute(f"SELECT name FROM sqlite_master WHERE type='table' AND name=?;", (table_name,))
+        table_exists = cursor.fetchone() is not None
+
+        if table_exists:
+            # Get existing columns and types
+            cursor.execute(f"PRAGMA table_info({table_name})")
+            existing_cols_info = cursor.fetchall()  # [(cid, name, type, notnull, dflt_value, pk), ...]
+            existing_cols = {col[1]: col[2].upper() for col in existing_cols_info}
+
+            # Automatically add missing columns with default empty string
+            missing_cols = [col for col in headers if col not in existing_cols]
+            if missing_cols:
+                first_row = data_list[0]
+                for col in missing_cols:
+                    index = headers.index(col)
+                    col_type = detect_type(first_row[index]) if auto_data_type else "TEXT"
+                    cursor.execute(
+                        f'ALTER TABLE {table_name} ADD COLUMN "{col}" {col_type} DEFAULT ""'
+                    )
+                    existing_cols[col] = col_type
+
+            # Coerce data to match existing types
+            if auto_data_type:
+                coerced_data = []
+                for row in data_list:
+                    new_row = []
+                    for val, col in zip(row, headers):
+                        col_type = existing_cols.get(col, "TEXT")
+                        if col_type == "INTEGER":
+                            new_row.append(int(val))
+                        elif col_type == "REAL":
+                            new_row.append(float(val))
+                        else:
+                            new_row.append(str(val))
+                    coerced_data.append(new_row)
+                data_list = coerced_data
+        else:
+            # Create table if it doesn't exist
+            first_row = data_list[0]
+            if auto_data_type:
+                columns_def = ", ".join([f'"{col}" {detect_type(val)} DEFAULT ""' for col, val in zip(headers, first_row)])
+            else:
+                columns_def = ", ".join([f'"{col}" TEXT DEFAULT ""' for col in headers])
+            cursor.execute(f"CREATE TABLE IF NOT EXISTS {table_name} ({columns_def})")
+
+        # Insert rows
+        placeholders = ", ".join(["?" for _ in headers])
+        cursor.executemany(f"INSERT INTO {table_name} ({', '.join(headers)}) VALUES ({placeholders})", data_list)
+
+        conn.commit()
+        conn.close()
+        # print(f"Saved {len(data_list)} rows to '{table_name}' in '{db_file_path}'.")
+
+    except Exception as e:
+        print(f"save_to_db: {e.__class__} - {str(e)}")
+
 # -------------------------------
 # File handling
 # -------------------------------
@@ -441,6 +628,7 @@ def save_file_content(output_file_path: str = None, content: Any = "", encoding:
 __all__ = [
     "create_directory",
     "standardized_string",
+    "replace_content",
     "get_base_domain",
     "read_csv",
     "save_to_csv",
